@@ -1,7 +1,7 @@
 ;;;
 ;;; http.scm - HTTP 1.1
 ;;;
-;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2019  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -54,6 +54,7 @@
   (use gauche.charconv)
   (use gauche.sequence)
   (use gauche.uvector)
+  (use gauche.connection)
   (use util.match)
   (use text.tree)
   (export <http-error>
@@ -82,9 +83,8 @@
           mime-compose-message-string
           mime-compose-parameters
           mime-parse-content-type)
-(autoload rfc.tls
-          make-tls tls-destroy tls-connect tls-input-port tls-output-port tls-close)
-
+(autoload rfc.tls make-tls tls-connect)
+(autoload rfc.http.tunnel probe-stunnel make-stunnel-connection)
 (autoload file.util file-size find-file-in-paths null-device)
 
 ;;==============================================================
@@ -106,7 +106,6 @@
 (define http-proxy (make-parameter #f))
 
 ;; The default redirect handler
-;;
 (define http-default-redirect-handler
   (make-parameter
    (^[method code headers body]
@@ -225,7 +224,6 @@
                            auth-user
                            auth-password
                            (proxy (http-proxy))
-                           extra-headers
                            (user-agent (http-user-agent))
                            (secure #f)
                            (receiver (http-string-receiver))
@@ -233,6 +231,12 @@
                            ((:request-encoding enc) (gauche-character-encoding))
                       :allow-other-keys opts)
 
+  (define extra-headers
+    ($ concatenate $ reverse
+       $ rlet1 z '()
+       (do-plist [(k v) opts]
+         (when v (push! z (list k v))))))
+
   (define conn (ensure-connection server auth-handler auth-user auth-password
                                   proxy secure extra-headers))
   (define redirector (if no-redirect
@@ -241,7 +245,6 @@
                          [(#t) (http-default-redirect-handler)]
                          [(#f) #f]
                          [else => identity])))
-  (define options `(:user-agent ,user-agent ,@(http-auth-headers conn) ,@opts))
   (define no-body-replies '("204" "304"))
 
   (define (get-body iport method code headers receiver)
@@ -255,16 +258,16 @@
                                         '(:proxy-connection keep-alive)
                                         '(:connection keep-alive))]
                [#t @ `(:host ,host :user-agent ,user-agent
-                       ,@(http-auth-headers conn) ,@opts)]))
+                       ,@(http-auth-headers conn) ,@extra-headers)]))
 
   ;; If we decide to give up redirection, we read from already-retrieved
   ;; body of 3xx reply.  This modifies reply headers if necessary.
   (define (redirect-headers body rep-headers)
     (if body
-      `(:content-length ,(string-size body)
-                        ,@(delete-keywords '(:content-length
-                                             :content-transfer-encoding)
-                                           rep-headers))
+      `(("content-length" ,(string-size body))
+        ,@(remove (^p (member (car p) '("content-length"
+                                        "content-transfer-encoding")))
+                  rep-headers))
       rep-headers))
 
   ;; returns either one of:
@@ -278,9 +281,11 @@
         (let* ([body (get-body in method code rep-headers
                                (http-string-receiver))]
                [verdict (consider-redirect method code rep-headers body)])
+          ;; consider-redirect returns either #f (don't redirect) or
+          ;; (METHOD . LOCATION).
           (if verdict
             `(redirect-to ,(car verdict) ,(cdr verdict))
-            (let1 hdrs (redirect-headers body rep-headers)
+            (let1 hdrs (redirect-headers body rep-headers) ;giving up
               `(reply ,code ,hdrs
                       ,(and body
                             (receive-body (open-input-string body) code
@@ -311,6 +316,7 @@
              (loop (cons uri history)
                    (~ (redirect-connection! conn proto new-server)'server)
                    method path*))])))))
+
 ;;
 ;; Pre-defined receivers
 ;;
@@ -487,18 +493,17 @@
 (define-class <http-connection> ()
   ;; All slots are private.
   ((server :init-keyword :server)       ; server[:port]
-   (socket :init-value #f)              ; A <socket> for persistent connection.
+   (socket :init-value #f)              ; A <connection> object; either <socket>
+                                        ; or wrapped by secure agent.
                                         ; If it is shutdown by the server,
                                         ; the APIs attempt to reconnect.
-   (secure-agent  :init-value #f)       ; When using secure connection via
-                                        ; tls, this slot holds its handle.
    (persistent    :init-keyword :persistent)   ; true for persistent connection.
    (auth-handler  :init-keyword :auth-handler  :init-value #f)
    (auth-user     :init-keyword :auth-user     :init-value #f)
    (auth-password :init-keyword :auth-password :init-value #f)
    (proxy         :init-keyword :proxy)
    (extra-headers :init-keyword :extra-headers)
-   (secure        :init-keyword :secure) ; boolean
+   (secure        :init-keyword :secure) ; either #f, tls or stunnel
    ))
 
 (define (make-http-connection server :key
@@ -520,17 +525,43 @@
 ;; This modifies CONN.
 (define (redirect-connection! conn proto new-server)
   (let1 orig-server (~ conn'server)
+    ;; Unless we're connecting the same server with same secure level,
+    ;; we reset the connection.
     (unless (and (string=? orig-server new-server)
-                 (eq? (~ conn'secure) (equal? proto "https")))
-      (shutdown-secure-agent conn)
-      (and-let* ([s (~ conn'socket)])
-        (socket-shutdown s)
-        (socket-close s)
+                 (or (and (equal? proto "https") (~ conn'secure))
+                     (and (not (equal? proto "https")) (not (~ conn'secure)))))
+      (and-let1 c (~ conn'socket)
+        (connection-shutdown (~ conn'socket) 'both)
         (set! (~ conn'socket) #f))
       (set! (~ conn'server) new-server)
-      (set! (~ conn'secure) (equal? proto "https"))))
+      ;; NB: If the server redirects http connection to https, we don't have
+      ;; a chance for the caller to select alternative TLS flavor---for the
+      ;; time being we blindly assume tls.
+      (set! (~ conn'secure) (and (equal? proto "https") 'tls))))
   conn)
 
+;; API
+(define (reset-http-connection conn)
+  (when (~ conn'socket)
+    (guard (e [(<system-error> e) #f])
+      (connection-shutdown (~ conn'socket) 'both)
+      (connection-close (~ conn'socket)))
+    (set! (~ conn'socket) #f)))
+
+
+;; API
+(define (http-secure-connection-available? :optional (type #t))
+  (case type
+    [(#t tls)
+     (cond-expand
+      [gauche.net.tls #t]
+      [else #f])]
+    [(stunnel)
+     (boolean (probe-stunnel))]
+    [else
+     (error "Unknown secure connection type (must be tls, stunnel or #t):"
+            type)]))
+
 ;;==============================================================
 ;; query and request body composition
 ;;
@@ -665,18 +696,6 @@
     [(path n&v ...) (http-compose-query path n&v enc)]
     [_ (error "Invalid request-uri form for http request API:" request-uri)]))
 
-(define (canonical-body request-body extra-headers enc)
-  (cond [(not request-body) (values #f extra-headers)]
-        [(string? request-body) (values request-body extra-headers)]
-        [(list? request-body)
-         (receive (body boundary) (http-compose-form-data request-body #f enc)
-           (values body
-                   `(:mime-version "1.0"
-                     :content-type
-                     ,#"multipart/form-data; boundary=~boundary"
-                     ,@(delete-keyword! :content-type extra-headers))))]
-        [else (error "Invalid request-body format:" request-body)]))
-
 ;; Always returns a connection object.
 (define (ensure-connection server auth-handler auth-user auth-password
                            proxy secure extra-headers)
@@ -694,13 +713,17 @@
       (check-override auth-password)
       (check-override proxy)
       (check-override extra-headers)
-      (check-override secure))))
+      (set! (~ conn'secure)
+            (and (not (undefined? secure))
+                 (case secure
+                   [(#f) #f]
+                   [(#t tls) 'tls]
+                   [(stunnel) 'stunnel]
+                   [else (error "`secure' argument for http connection must be \
+                                 either boolean, 'tls or 'stunnel, but got:"
+                                secure)]))))))
 
-(define (reset-http-connection conn)
-  (shutdown-socket-connection conn)
-  (shutdown-secure-agent conn))
-
-(define (start-socket-connection conn)
+(define (start-connection conn)
   ;; If address is given ipv6 format such as "[::1]:port", we have to
   ;; use "::1" part as the hostname, excluding [].
   (define (parse-address addr)
@@ -709,34 +732,34 @@
       [#/^\[([a-fA-F\d:]+)\](?::(\d+))?$/ (_ host port) (values host port #f)]
       [#/^([^:]+)(?::(\d+))?$/ (_ host port) (values host port #f)]
       [else (error "Unrecognized http server address:" addr)]))
-  (receive (host port path) (parse-address (or (~ conn'proxy) (~ conn'server)))
-    (set! (~ conn'socket)
-          (if path
-            (make-client-socket 'unix path)
-            (make-client-socket host (if port
-                                       (x->integer port)
-                                       (if (~ conn'secure) 443 80)))))))
-
-(define (shutdown-socket-connection conn)
-  (when (~ conn'socket)
-    (guard (e [(<system-error> e) #f])
-      (socket-shutdown (~ conn'socket)))
-    (socket-close (~ conn'socket))
-    (set! (~ conn'socket) #f)))
+  (define (connect-socket)
+    (receive (host port path) (parse-address (or (~ conn'proxy) (~ conn'server)))
+      (if path
+        (make-client-socket 'unix path)
+        (make-client-socket host (if port
+                                   (x->integer port)
+                                   (if (~ conn'secure) 443 80))))))
+  (ecase (~ conn'secure)
+    [(#f) (set! (~ conn'socket) (connect-socket))]
+    [(tls)
+     (let1 tls (make-tls :server-name (~ conn'server)
+                         :options (if (tls-ca-bundle-path)
+                                    0
+                                    SSL_SERVER_VERIFY_LATER))
+       (set! (~ conn'socket) (tls-connect tls (connect-socket))))]
+    [(stunnel)
+     (let* ([rhost      (or (~ conn'proxy) (~ conn'server))]
+            [rhost:port (if (string-index rhost #\:) rhost #"~|rhost|:https")])
+       (set! (~ conn'socket) (make-stunnel-connection rhost:port)))]))
 
 (define (with-connection conn proc)
-  (unless (~ conn'persistent)
-    (unless (~ conn'socket) (start-socket-connection conn))
-    (when (~ conn'secure) (start-secure-agent conn)))
   (unwind-protect
-      (apply proc (if (~ conn'secure)
-                    `(,(tls-input-port (~ conn'secure-agent))
-                      ,(tls-output-port (~ conn'secure-agent)))
-                    `(,(socket-input-port (~ conn'socket))
-                      ,(socket-output-port (~ conn'socket)))))
+      (begin
+        (unless (~ conn'socket) (start-connection conn))
+        (proc (connection-input-port (~ conn'socket))
+              (connection-output-port (~ conn'socket))))
     (unless (~ conn'persistent)
-      (when (~ conn'secure) (shutdown-secure-agent conn))
-      (shutdown-socket-connection conn))))
+      (reset-http-connection conn))))
 
 ;; canonicalize uri for the sake of redirection.
 ;; URI is a request-uri given to the API, or the redirect location specified
@@ -859,30 +882,6 @@
   (begin0 (receiver code headers total callback)
     (when condition (raise condition))))
 
-;;==============================================================
-;; secure agent handling
-;;
-
-(define (shutdown-secure-agent conn)
-  (when (~ conn'secure-agent)
-    (tls-close (~ conn'secure-agent))
-    (tls-destroy (~ conn'secure-agent))
-    (set! (~ conn'secure-agent) #f)))
-
-(define (start-secure-agent conn)
-  (unless (http-secure-connection-available?)
-    (error "Secure connection is not available on this platform"))
-  (when (~ conn'secure-agent) (shutdown-secure-agent conn))
-  (let1 tls (make-tls)
-    (tls-connect tls (socket-fd (~ conn'socket)))
-    (set! (~ conn'secure-agent) tls)))
-
-;; for external api
-(define (http-secure-connection-available?)
-  (cond-expand
-   [gauche.net.tls #t]
-   [else #f]))
-
 ;;==============================================================
 ;; authentication handling
 ;;
